#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module for the NaiveKDE.
"""
import itertools
import numbers

import numpy as np

from KDEpy.BaseKDE import BaseKDE


class NaiveKDE(BaseKDE):
    """
    This class implements a naive computation of a kernel density estimate. The
    advantages are that choices of bandwidth, norms, weights and grids are
    straightforward -- the user can do almost anything. The disadvantage is
    that computations are slow on more than a couple of thousand data points.

    Parameters
    ----------
    kernel : str
        The kernel function. See cls._available_kernels.keys() for choices.
    bw : float, str or array-like
        Bandwidth or bandwidth selection method. If a float is passed, it
        is the standard deviation of the kernel. If a string it passed, it
        is the bandwidth selection method, see cls._bw_methods.keys() for
        choices. If an array-like it passed, it is the bandwidth of each
        point.
    norm : float
        The p-norm used to compute the distances in higher dimensions.

    Examples
    --------
    >>> data = np.random.randn(2**10)
    >>> # (1) Automatic bw selection using Improved Sheather Jones
    >>> x, y = NaiveKDE(bw='ISJ').fit(data).evaluate()
    >>> # (2) Explicit choice of kernel and bw (standard deviation of kernel)
    >>> x, y = NaiveKDE(kernel='triweight', bw=0.5).fit(data).evaluate()
    >>> weights = data + 10
    >>> # (3) Using a grid and weights for the data
    >>> y = NaiveKDE(kernel='epa', bw=0.5).fit(data, weights).evaluate(x)

    References
    ----------
    - Silverman, B. W. Density Estimation for Statistics and Data Analysis.
      Boca Raton: Chapman and Hall, 1986.
    - Wand, M. P., and M. C. Jones. Kernel Smoothing.
      London ; New York: Chapman and Hall/CRC, 1995.
    - Scipy implementation, at ``scipy.stats.gaussian_kde``.
    """

    def __init__(self, kernel="gaussian", bw=1, norm=2):
        super().__init__(kernel, bw)
        self.norm = norm

    def fit(self, data, weights=None):
        """
        Fit the KDE to the data. This validates the data and stores it.
        Computations are performed when the KDE is evaluated on a grid.

        Parameters
        ----------
        data: array-like
            The data points. High dimensional data must have shape (obs, dims).
        weights: array-like
            One weight per data point. Must have shape (obs,). If None is
            passed, uniform weights are used.

        Returns
        -------
        self
            Returns the instance.

        Examples
        --------
        >>> data = [1, 3, 4, 7]
        >>> weights = [3, 4, 2, 1]
        >>> kde = NaiveKDE().fit(data, weights=None)
        >>> kde = NaiveKDE().fit(data, weights=weights)
        >>> x, y = kde()
        """
        # Sets self.data
        super().fit(data, weights)
        return self

    def evaluate(self, grid_points=None):
        """
        Evaluate on grid points.

        Parameters
        ----------
        grid_points: array-like, int, tuple or None
            A grid (mesh) to evaluate on. High dimensional grids must have
            shape (obs, dims). If an integer is passed, it's the number of grid
            points on an equidistant grid. If a tuple is passed, it's the
            number of grid points in each dimension. If None, a grid will be
            automatically created.

        Returns
        -------
        y: array-like
            If a grid is supplied, `y` is returned. If no grid is supplied,
            a tuple (`x`, `y`) is returned.

        Examples
        --------
        >>> kde = NaiveKDE().fit([1, 3, 4, 7])
        >>> # Two ways to evaluate, either with a grid or without
        >>> x, y = kde.evaluate()
        >>> x, y = kde.evaluate(256)
        >>> y = kde.evaluate(x)
        """
        # This method sets self.grid points and verifies it
        # NaiveKDE does not convert the bw to a scalar, since a vector is
        # allowed too.
        super().evaluate(grid_points, bw_to_scalar=False)

        # Create zeros on the grid points
        evaluated = np.zeros(self.grid_points.shape[0])

        # For every data point, compute the kernel and add to the grid
        bw = self.bw
        if isinstance(bw, numbers.Number):
            bw = np.asarray(np.ones(self.data.shape[0]) * bw, dtype=float)

        # TODO: Implementation w.r.t grid points for faster evaluation
        # See the SciPy evaluation for how this can be done
        weights = itertools.repeat(1 / self.data.shape[0]) if self.weights is None else self.weights

        for weight, data_point, bw in zip(weights, self.data, bw):
            x = self.grid_points - data_point
            evaluated += weight * self.kernel(x, bw=bw, norm=self.norm)

        return self._evalate_return_logic(evaluated, self.grid_points)


if __name__ == "__main__":
    import pytest

    # --durations=10  <- May be used to show potentially slow tests
    pytest.main(args=[".", "--doctest-modules", "-v"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import importlib.metadata


from KDEpy.FFTKDE import FFTKDE
from KDEpy.NaiveKDE import NaiveKDE
from KDEpy.TreeKDE import TreeKDE


__version__ = importlib.metadata.version("KDEpy")
__author__ = "tommyod"

__all__ = ["TreeKDE", "NaiveKDE", "FFTKDE"]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module for functions related to linear binning. These functions perform
linear binning on weighted data. This is typically a preprocessing step
before convolving with a kernel in the FFTKDE, but may also be used to
reduce the effective number of data points in any algorithm.

The idea behind linear binning is the following: (1) go through every
data point and (2) assign a weight to the 2^dims nearest grid points.
In `dims` dimensions, there are 2 points on the grid to consider in
each dimension, so a total of 2^dims grid points must be assigned weights to
for every data point. The weights are determined by the proportion of
the volume of this hypercube that is enclosed by the data point.

A ------------------------------ B
|                    |           |
|                    |           |
|                    X-----------|
|                                |
|                                |
|                                |
C ------------------------------ C

References
----------
- Fan, Jianqing, and James S. Marron.
  “Fast Implementations of Nonparametric Curve Estimators.”
  Journal of Computational and Graphical Statistics 3, no. 1 (March 1, 1994).
  https://doi.org/10.1080/10618600.1994.10474629.
"""
import functools
import itertools
import operator

import numpy as np

# Import compiled _cutils module (not declared in __init__)
# and ignore unresolved reference (may not be build in dev environment)
# noinspection PyProtectedMember, PyUnresolvedReferences
from KDEpy import _cutils
from KDEpy.utils import cartesian


# This parameter was in use when a NumPy implementation and a Cython
# implementation were both used. Now only Cython is used.
_use_Cython = True


def linbin_cython(data, grid_points, weights=None):
    """
    1D Linear binning using Cython. Assigns weights to grid points from data.

    Runs in approx 10 ms on 1 million data points.

    Parameters
    ----------
    data : array-like
        The data to bin. Must be of shape (obs,).
    grid_points : array-like
        Equidistant grid points to assign weights to.
        Must be of shape (points,).
    weights : array-like
        The weights of the data points.
        Must be of shape (obs,).

    Examples
    --------
    >>> data = np.array([2, 2.5, 3, 4])
    >>> ans = linbin_cython(data, np.arange(6), weights=None)
    >>> np.allclose(ans, np.array([0, 0, 0.375, 0.375, 0.25, 0]))
    True
    >>> ans = linbin_cython(data, np.arange(6), weights=np.arange(1, 5))
    >>> np.allclose(ans, np.array([0, 0, 0.2, 0.4, 0.4, 0]))
    True
    >>> data = np.array([2, 2.5, 3, 4])
    >>> ans = linbin_cython(data, np.arange(1, 7), weights=None)
    >>> np.allclose(ans, np.array([0, 0.375, 0.375, 0.25, 0, 0]))
    True
    """
    # Convert the data and grid points
    data = np.asarray_chkfinite(data, dtype=float)
    grid_points = np.asarray_chkfinite(grid_points, dtype=float)

    assert len(data.shape) == 1
    assert len(grid_points.shape) == 1

    # Verify that the grid is equidistant
    diffs = np.diff(grid_points)
    assert np.allclose(np.ones_like(diffs) * diffs[0], diffs)

    if weights is not None:
        assert len(weights.shape) == 1
        weights = np.asarray_chkfinite(weights, dtype=float)
        weights = weights / np.sum(weights)

    if (weights is not None) and (len(data) != len(weights)):
        raise ValueError("Length of data must match length of weights.")

    # Transform the data
    min_grid = np.min(grid_points)
    max_grid = np.max(grid_points)
    num_intervals = len(grid_points) - 1  # Number of intervals
    dx = (max_grid - min_grid) / num_intervals
    transformed_data = (data - min_grid) / dx

    result = np.asarray(np.zeros(num_intervals + 2), dtype=float)

    # Two Cython functions are implemented, one for weighted data and one
    # for unweighted data, since creating equal weights is costly w.r.t time
    if weights is None:
        result = _cutils.iterate_data_1D(transformed_data, result)
        return np.asarray(result[:-1], dtype=float) / transformed_data.shape[0]
    else:
        res = _cutils.iterate_data_1D_weighted(transformed_data, weights, result)
        return np.asarray(res[:-1], dtype=float)  # Remove last, outside of grid


def linbin_numpy(data, grid_points, weights=None):
    """
    1D Linear binning using NumPy. Assigns weights to grid points from data.

    This function is fast for data sets upto approximately 1-10 million,
    it uses vectorized NumPy functions to perform linear binning. Takes around
    100 ms on 1 million data points, so not nearly as fast as the Cython
    implementation (10 ms).

    Parameters
    ----------
    data : array-like
        Must be of shape (obs,).
    grid_points : array-like
        Must be of shape (points,).
    weights : array-like
        Must be of shape (obs,).

    Examples
    --------
    >>> data = np.array([2, 2.5, 3, 4])
    >>> ans = linbin_numpy(data, np.arange(6), weights=None)
    >>> np.allclose(ans, np.array([0, 0, 0.375, 0.375, 0.25, 0]))
    True
    >>> ans = linbin_numpy(data, np.arange(6), weights=np.arange(1, 5))
    >>> np.allclose(ans, np.array([0, 0, 0.2, 0.4, 0.4, 0]))
    True
    >>> data = np.array([2, 2.5, 3, 4])
    >>> ans = linbin_numpy(data, np.arange(1, 7), weights=None)
    >>> np.allclose(ans, np.array([0, 0.375, 0.375, 0.25, 0, 0]))
    True
    """
    # Convert the data and grid points
    data = np.asarray_chkfinite(data, dtype=float)
    grid_points = np.asarray_chkfinite(grid_points, dtype=float)
    assert len(data.shape) == 1
    assert len(grid_points.shape) == 1

    # Verify that the grid is equidistant
    diffs = np.diff(grid_points)
    assert np.allclose(np.ones_like(diffs) * diffs[0], diffs)

    if weights is None:
        weights = np.ones_like(data)

    weights = np.asarray_chkfinite(weights, dtype=float)
    weights = weights / np.sum(weights)

    if not len(data) == len(weights):
        raise ValueError("Length of data must match length of weights.")

    # Transform the data
    min_grid = np.min(grid_points)
    max_grid = np.max(grid_points)
    num_intervals = len(grid_points) - 1
    dx = (max_grid - min_grid) / num_intervals
    transformed_data = (data - min_grid) / dx

    # Compute the integral and fractional part of the data
    # The integral part is used for lookups, the fractional part is used
    # to weight the data
    fractional, integral = np.modf(transformed_data)
    integral = integral.astype(int)

    # Sort the integral values, and the fractional data and weights by
    # the same key. This lets us use binary search, which is faster
    # than using a mask in the the loop below
    indices_sorted = np.argsort(integral)
    integral = integral[indices_sorted]
    fractional = fractional[indices_sorted]
    weights = weights[indices_sorted]

    # Pre-compute these products, as they are used in the loop many times
    frac_weights = fractional * weights
    neg_frac_weights = weights - frac_weights

    # If the data is not a subset of the grid, the integral values will be
    # outside of the grid. To solve the problem, we filter these values away
    unique_integrals = np.unique(integral)
    unique_integrals = unique_integrals[(unique_integrals >= 0) & (unique_integrals <= len(grid_points))]

    result = np.asarray(np.zeros(len(grid_points) + 1), dtype=float)
    for grid_point in unique_integrals:
        # Use binary search to find indices for the grid point
        # Then sum the data assigned to that grid point
        low_index = np.searchsorted(integral, grid_point, side="left")
        high_index = np.searchsorted(integral, grid_point, side="right")
        result[grid_point] += neg_frac_weights[low_index:high_index].sum()
        result[grid_point + 1] += frac_weights[low_index:high_index].sum()

    return result[:-1]


def linbin_Ndim_python(data, grid_points, weights=None):
    """
    d-dimensional linear binning. This is a slow, pure-Python function.
    Mainly used for testing purposes.

    With :math:`N` data points, and :math:`n` grid points in each dimension
    :math:`d`, the running time is :math:`O(N2^d)`. For each point the
    algorithm finds the nearest points, of which there are two in each
    dimension.

    Parameters
    ----------
    data : array-like
        The data must be of shape (obs, dims).
    grid_points : array-like
        Grid, where cartesian product is already performed.
    weights : array-like
        Must have shape (obs,).

    Examples
    --------
    >>> from KDEpy.utils import autogrid
    >>> grid_points = autogrid(np.array([[0, 0, 0]]), num_points=(3, 3, 3))
    >>> d = linbin_Ndim_python(np.array([[1.0, 0, 0]]), grid_points, None)
    """
    # Convert the data and grid points
    data = np.asarray_chkfinite(data, dtype=float)
    grid_points = np.asarray_chkfinite(grid_points, dtype=float)
    if weights is not None:
        weights = np.asarray_chkfinite(weights, dtype=float)
    else:
        # This is not efficient, but this function should just be correct
        # The faster algorithm is implemented in Cython
        weights = np.ones(data.shape[0])
    weights = weights / np.sum(weights)

    if (weights is not None) and (data.shape[0] != len(weights)):
        raise ValueError("Length of data must match length of weights.")

    obs_tot, dims = grid_points.shape

    # Compute the number of grid points for each dimension in the grid
    grid_num = (grid_points[:, i] for i in range(dims))
    grid_num = np.array(list(len(np.unique(g)) for g in grid_num))

    # Scale the data to the grid
    min_grid = np.min(grid_points, axis=0)
    max_grid = np.max(grid_points, axis=0)
    num_intervals = grid_num - 1  # Number of intervals
    dx = (max_grid - min_grid) / num_intervals
    data = (data - min_grid) / dx

    # Create results
    result = np.zeros(grid_points.shape[0], dtype=float)

    # Go through every data point
    for observation, weight in zip(data, weights):
        # Compute integer part and fractional part for every x_i
        # Compute relation to previous grid point, and next grid point
        int_frac = (
            (
                (int(coordinate), 1 - (coordinate % 1)),
                (int(coordinate) + 1, (coordinate % 1)),
            )
            for coordinate in observation
        )

        # Go through every cartesian product, i.e. every corner in the
        # hypercube grid points surrounding the observation
        for cart_prod in itertools.product(*int_frac):
            fractions = (frac for (integral, frac) in cart_prod)
            integrals = list(integral for (integral, frac) in cart_prod)
            # Find the index in the resulting array, compured by
            # x_1 * (g_2 * g_3 * g_4) + x_2 * (g_3 * g_4) + x_3 * (g_4) + x_4

            index = integrals[0]
            for j in range(1, dims):
                index = grid_num[j] * index + integrals[j]

            value = functools.reduce(operator.mul, fractions)
            result[index % obs_tot] += value * weight

    assert np.allclose(np.sum(result), 1)
    return result


def linbin_Ndim(data, grid_points, weights=None):
    """
    d-dimensional linear binning, when d >= 2.

    With :math:`N` data points, and :math:`n` grid points in each dimension
    :math:`d`, the running time is :math:`O(N2^d)`. For each point the
    algorithm finds the nearest points, of which there are two in each
    dimension. Approximately 200 times faster than pure Python implementation.

    Parameters
    ----------
    data : array-like
        The data must be of shape (obs, dims).
    grid_points : array-like
        Grid, where cartesian product is already performed.
    weights : array-like
        Must have shape (obs,).

    Examples
    --------
    >>> from KDEpy.utils import autogrid
    >>> grid_points = autogrid(np.array([[0, 0, 0]]), num_points=(3, 3, 3))
    >>> d = linbin_Ndim(np.array([[1.0, 0, 0]]), grid_points, None)
    """
    data_obs, data_dims = data.shape
    assert len(grid_points.shape) == 2
    assert data_dims >= 2

    # Convert the data and grid points
    data = np.asarray_chkfinite(data, dtype=float)
    grid_points = np.asarray_chkfinite(grid_points, dtype=float)
    if weights is not None:
        weights = np.asarray_chkfinite(weights, dtype=float)
        weights = weights / np.sum(weights)

    if (weights is not None) and (data.shape[0] != len(weights)):
        raise ValueError("Length of data must match length of weights.")

    obs_tot, dims = grid_points.shape

    # Compute the number of grid points for each dimension in the grid
    grid_num = (grid_points[:, i] for i in range(dims))
    grid_num = np.array(list(len(np.unique(g)) for g in grid_num), dtype="long")

    # Scale the data to the grid
    min_grid = np.min(grid_points, axis=0)
    max_grid = np.max(grid_points, axis=0)
    num_intervals = grid_num - 1
    dx = (max_grid - min_grid) / num_intervals
    data = (data - min_grid) / dx

    # Create results
    result = np.zeros(grid_points.shape[0], dtype=float)

    # Call the Cython implementation. Loops are unrolled if d=1 or d=2,
    # and if d >= 3 a more general routine is called. It's a bit slower since
    # the loops are not unrolled.

    # Weighted data has two specific routines
    if weights is not None:
        if data_dims >= 3:
            binary_flgs = cartesian(([0, 1],) * dims).astype("long")
            result = _cutils.iterate_data_ND_weighted(data, weights, result, grid_num, obs_tot, binary_flgs)
        else:
            result = _cutils.iterate_data_2D_weighted(data, weights, result, grid_num, obs_tot)
        result = np.asarray_chkfinite(result, dtype=float)

    # Unweighted data has two specific routines too. This is because creating
    # uniform weights takes relatively long time. It's faster to have a
    # specialize routine for this case.
    else:
        if data_dims >= 3:
            binary_flgs = cartesian(([0, 1],) * dims).astype("long")
            result = _cutils.iterate_data_ND(data, result, grid_num, obs_tot, binary_flgs)
        else:
            result = _cutils.iterate_data_2D(data, result, grid_num, obs_tot)
        result = np.asarray_chkfinite(result, dtype=float)
        result = result / data_obs

    assert np.allclose(np.sum(result), 1)
    return result


def linear_binning(data, grid_points, weights=None):
    """
    This wrapper function computes d-dimensional binning, very quickly.

    Computes binning by setting a linear grid and weighting points linearily
    by their distance to the grid points. In addition, weight asssociated with
    data points may be passed. Depending on whether or not weights are passed
    and the dimensionality of the data, specific sub-routines are called for
    fast evaluation.

    Parameters
    ----------
    data
        The data points.
    grid_points
        The number of points in the grid.
    weights
        The weights.

    Returns
    -------
    (grid, data)
        Data weighted at each grid point.

    Examples
    --------
    >>> data = [1, 1.5, 1.5, 2, 2.8, 3]
    >>> grid_points = [1, 2, 3]
    >>> data = linear_binning(data, grid_points)
    >>> np.allclose(data, np.array([0.33333, 0.36667, 0.3]))
    True
    """
    data = np.asarray_chkfinite(data, dtype=float)
    grid_points = np.asarray_chkfinite(grid_points, dtype=float)
    if weights is not None:
        weights = np.asarray_chkfinite(weights, dtype=float)

    # Make sure the dimensionality makes sense
    try:
        data_obs, data_dims = data.shape
    except ValueError:
        data_dims = 1

    try:
        grid_obs, grid_dims = grid_points.shape
    except ValueError:
        grid_dims = 1

    if not data_dims == grid_dims:
        raise ValueError("Shape of data and grid points must be the same.")

    if data_dims == 1:
        if _use_Cython:
            return linbin_cython(data.ravel(), grid_points.ravel(), weights=weights)
        else:
            return linbin_numpy(data.ravel(), grid_points.ravel(), weights=weights)
    else:
        return linbin_Ndim(data, grid_points, weights=weights)


if __name__ == "__main__":
    import pytest

    # --durations=10  <- May be used to show potentially slow tests
    pytest.main(args=[".", "--doctest-modules", "-v", "--capture=sys"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module for kernel functions. A kernel function is a radial basis function which
is everywhere non-negative and whose integral evalutes to unity. Every kernel
function takes an `x` of shape (obs, dims) and returns a y of shape (obs, 1).
"""

import collections.abc
import functools
import numbers

import numpy as np
from scipy.optimize import brentq
from scipy.special import factorial2, gamma
from scipy.stats import norm

# In R, the following are implemented:
# "gaussian", "rectangular", "triangular", "epanechnikov",
# "biweight", "cosine" or "optcosine"

# Wikipedia
# uniform, trinagular, epanechnikov, quartic,triweight, tricube,
# gaussian, cosine, logistic, sigmoid, silverman

# All kernel functions take x of shape (obs, dims) and returns (obs, 1)
# All kernel functions integrate to unity


def gauss_integral(n):
    r"""
    Solve the integral
    \int_0^1 exp(-0.5 * x * x) x^n dx

    See
    https://en.wikipedia.org/wiki/List_of_integrals_of_Gaussian_functions

    Examples
    --------
    >>> ans = gauss_integral(3)
    >>> np.allclose(ans, 2)
    True
    >>> ans = gauss_integral(4)
    >>> np.allclose(ans, 3.75994)
    True
    """
    factor = np.sqrt(np.pi * 2)

    # Special case:
    # factorial2(-1) used to return 1
    # but now it returns 0
    # https://github.com/scipy/scipy/pull/15841
    # https://github.com/tommyod/KDEpy/issues/145
    if n == 0:
        return factor * factorial2(n - 0) / 2

    # Normal cases
    elif n % 2 == 0:
        return factor * factorial2(n - 1) / 2
    elif n % 2 == 1:
        return factor * norm.pdf(0) * factorial2(n - 1)
    else:
        raise ValueError("n must be odd or even.")


def trig_integral(k):
    """
    Returns the solutions to
    Is(k) = int_0^1 sin(pi * x / 2) x^k dx
    Ic(k) = int_0^1 cos(pi * x / 2) x^k dx
    using a recursive formula. Returns a tuple (Is, Ic).

    Examples
    --------
    >>> import numpy as np
    >>> Is, Ic = trig_integral(2) # Verify with solution from WolframAlpha
    >>> np.allclose([Is, Ic], [0.29454, 0.12060], rtol=10e-5)
    True
    """

    Ic = 2 / np.pi
    Is = 2 / np.pi

    if k <= 0:
        return Is, Ic

    for i in range(1, k + 1):
        Ic, Is = (2 / np.pi) * (1 - i * Is), (2 / np.pi) * i * Ic

    return Is, Ic


def p_norm(x, p):
    """
    The p-norm of an array of shape (obs, dims)

    Examples
    --------
    >>> x = np.arange(9).reshape((3, 3))
    >>> p = 2
    >>> np.allclose(p_norm(x, p), euclidean_norm(x))
    True
    """
    if np.isinf(p):
        return infinity_norm(x)
    elif p == 2:
        return euclidean_norm(x)
    elif p == 1:
        return taxicab_norm(x)
    return np.power(np.power(np.abs(x), p).sum(axis=1), 1 / p)


def euclidean_norm(x):
    """
    The 2 norm of an array of shape (obs, dims)
    """
    return np.sqrt((x * x).sum(axis=1))


def euclidean_norm_sq(x):
    """
    The squared 2 norm of an array of shape (obs, dims)
    """
    return (x * x).sum(axis=1)


def infinity_norm(x):
    """
    The infinity norm of an array of shape (obs, dims)
    """
    return np.abs(x).max(axis=1)


def taxicab_norm(x):
    """
    The taxicab norm of an array of shape (obs, dims)
    """
    return np.abs(x).sum(axis=1)


def volume_unit_ball(d, p=2):
    """
    Volume of d-dimensional unit ball under the p-norm. When p=1 this is called
    a cross-polytype, when p=2 it's called a hypersphere, and when p=infty it's
    called a hypercube.

    Notes
    -----
    See the following paper for a very general result related to this:

    - Wang, Xianfu. “Volumes of Generalized Unit Balls.”
      Mathematics Magazine 78, no. 5 (2005): 390–95.
      https://doi.org/10.2307/30044198.
    """
    return 2.0**d * gamma(1 + 1 / p) ** d / gamma(1 + d / p)


def epanechnikov(x, dims=1):
    normalization = 2 / (dims + 2)
    dist_sq = x**2
    out = np.zeros_like(dist_sq)
    mask = dist_sq < 1
    out[mask] = (1 - dist_sq)[mask] / normalization
    return out


def gaussian(x, dims=1):
    normalization = dims * gauss_integral(dims - 1)
    dist_sq = x**2
    return np.exp(-dist_sq / 2) / normalization


def box(x, dims=1):
    normalization = 1
    out = np.zeros_like(x)
    mask = x < 1
    out[mask] = 1 / normalization
    return out


def exponential(x, dims=1):
    normalization = gamma(dims) * dims
    return np.exp(-x) / normalization


def tri(x, dims=1):
    normalization = 1 / (dims + 1)
    out = np.zeros_like(x)
    mask = x < 1
    out[mask] = np.maximum(0, 1 - x)[mask] / normalization
    return out


def biweight(x, dims=1):
    normalization = 8 / ((dims + 2) * (dims + 4))
    dist_sq = x**2
    out = np.zeros_like(dist_sq)
    mask = dist_sq < 1
    out[mask] = np.maximum(0, (1 - dist_sq) ** 2)[mask] / normalization
    return out


def triweight(x, dims=1):
    normalization = 48 / ((dims + 2) * (dims + 4) * (dims + 6))
    dist_sq = x**2
    out = np.zeros_like(dist_sq)
    mask = dist_sq < 1
    out[mask] = np.maximum(0, (1 - dist_sq) ** 3)[mask] / normalization
    return out


def tricube(x, dims=1):
    normalization = 162 / ((dims + 3) * (dims + 6) * (dims + 9))
    out = np.zeros_like(x)
    mask = x < 1
    out[mask] = np.maximum(0, (1 - x**3) ** 3)[mask] / normalization
    return out


def cosine(x, dims=1):
    Is, Ic = trig_integral(dims - 1)
    normalization = Ic
    out = np.zeros_like(x)
    mask = x < 1
    out[mask] = np.cos((np.pi * x) / 2)[mask] / (normalization * dims)
    return out


def logistic(x, dims=1):
    return 1 / (2 + 2 * np.cosh(x))


def sigmoid(x, dims=1):
    return 1 / (np.pi * np.cosh(x))


class Kernel(collections.abc.Callable):
    def __init__(self, function, var=1, support=3):
        """
        Initialize a new kernel function.

        function: callable, numpy.arr -> numpy.arr, should integrate to 1
        expected_value : peak, typically 0
        support: support of the function.

        Example
        -------
        >>> from scipy.special import gamma
        >>> # Normalized function of x
        >>> def exp(x, dims=1):
        ...     normalization = gamma(dims) * dims
        ...     return np.exp(-x) / normalization
        >>> kernel = Kernel(exp, var=4, support=np.inf)
        >>> # The function is scaled so that the standard deviation (bw) = 1
        >>> assert kernel(0, bw=1, norm=2)[0] > kernel(1, bw=1, norm=2)[0]
        >>> np.allclose(kernel(np.array([0, 1, 2])), kernel([0, 1, 2]))
        True
        >>> np.allclose(kernel(0), kernel([0]))
        True
        >>> np.allclose(kernel(0), kernel.evaluate(0))
        True
        """
        self.function = function
        self.var = var
        self.finite_support = np.isfinite(support)

        # If the function has finite support, scale the support so that it
        # corresponds to the support of the function when it is scaled to have
        # unit variance.
        self.support = support / np.sqrt(self.var)

    def practical_support(self, bw, atol=10e-5):
        """
        Return the support for practical purposes. Used to find a support value
        for computations for kernel functions without finite (bounded) support.

        Examples
        --------
        >>> kernel = Kernel(gaussian, var=1, support=np.inf)
        >>> kernel.practical_support(bw=1)
        3.8994...
        >>> kernel.practical_support(bw=2)
        7.4331...
        """
        # If the kernel has finite support, return the support accounting for
        # the bw
        if self.finite_support:
            return self.support * bw

        # If the function does not have finite support, find a practical value
        else:

            def f(x):
                return self.evaluate(x, bw=bw)[0] - atol

            try:
                xtol = 1e-3
                ans = brentq(f, a=0, b=8 * bw, xtol=xtol, full_output=False)
                return ans + xtol
            except ValueError:
                msg = (
                    "Unable to solve for support numerically. Use a "
                    + "kernel with finite support or scale data to smaller bw."
                )
                raise ValueError(msg)

    def evaluate(self, x, bw=1, norm=2):
        """
        Evaluate the kernel.

        Parameters
        ----------
        x : array-like
            Should have shape (obs, dims).
        bw : array-like
            Must have shape (obs, ), or float.
        """

        # If x is a number, convert it to a length-1 NumPy vector
        if isinstance(x, numbers.Number):
            x = np.asarray_chkfinite([x])
        else:
            x = np.asarray_chkfinite(x)

        if len(x.shape) == 1:
            x = x.reshape(-1, 1)

        # Scale the function, such that bw=1 corresponds to the function having
        # a standard deviation (or variance) equal to 1
        real_bw = bw / np.sqrt(self.var)
        obs, dims = x.shape

        # Set the volume function
        volume_func = functools.partial(volume_unit_ball, p=norm)

        if dims > 1:
            distances = p_norm(x, norm).ravel()
        else:
            distances = np.abs(x).ravel()

        return self.function(distances / real_bw, dims) / ((real_bw**dims) * volume_func(dims))

    __call__ = evaluate


gaussian = Kernel(gaussian, var=1, support=np.inf)
exp = Kernel(exponential, var=2, support=np.inf)
box = Kernel(box, var=1 / 3, support=1)
tri = Kernel(tri, var=1 / 6, support=1)
epa = Kernel(epanechnikov, var=1 / 5, support=1)
biweight = Kernel(biweight, var=1 / 7, support=1)
triweight = Kernel(triweight, var=1 / 9, support=1)
tricube = Kernel(tricube, var=35 / 243, support=1)
cosine = Kernel(cosine, var=(1 - (8 / np.pi**2)), support=1)
logistic = Kernel(logistic, var=(np.pi**2 / 3), support=np.inf)
sigmoid = Kernel(sigmoid, var=(np.pi**2 / 4), support=np.inf)

_kernel_functions = {
    "gaussian": gaussian,
    "exponential": exp,
    "box": box,
    "tri": tri,
    "epa": epa,
    "biweight": biweight,
    "triweight": triweight,
    "tricube": tricube,
    "cosine": cosine,
    # 'logistic': logistic,
    # 'sigmoid': sigmoid
}


if __name__ == "__main__":
    import pytest

    # --durations=10  <- May be used to show potentially slow tests
    pytest.main(args=[__file__, "--doctest-modules", "-v"])



#!/usr/bin/env python3

"""KDEpy installer script using setuptools.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

import numpy as np
from Cython.Distutils import build_ext
from setuptools import Extension, setup

setup(
    packages=["KDEpy"],
    cmdclass={"build_ext": build_ext},
    include_dirs=[np.get_include()],
    ext_modules=[Extension("KDEpy._cutils", ["KDEpy/cutils_ext/cutils.pyx"])],
)
